from __future__ import print_function
import time
import requests
import json

from plantpredict.utilities import convert_json, camel_to_snake


def handle_refused_connection(function):
    def function_wrapper(*args, **kwargs):
        connection_error = True
        while connection_error:
            try:
                connection_error = False
                return function(*args, **kwargs)
            except requests.exceptions.ConnectionError:
                print("Connection refused, trying again...")
                time.sleep(7)
    function_wrapper.__name__ = function.__name__
    function_wrapper.__doc__ = function.__doc__
    return function_wrapper


def handle_error_response(function):
    def function_wrapper(*args, **kwargs):
        response = function(*args, **kwargs)
        try:
            # if the authorization is invalid, refresh the API access token
            if response.status_code == 401:
                args[0].api.__get_access_token()

            # if there is a sever side error, return the error message
            elif not 200 <= response.status_code < 300:
                raise APIError(response.status_code, response.content, response.url)

            # if the HTTP request receives a successful response
            else:

                # if the response contains content, return it
                if response.content:
                    if "Queue" in response.url:
                        return response.json()

                    else:
                        # if it is a list, use convert_json method in list comprehension
                        if isinstance(response.json(), list):
                            return [convert_json(i, camel_to_snake) for i in response.json()]
                        else:
                            return convert_json(response.json(), camel_to_snake)

                # if the response does not contain content, return a generic success message
                else:
                    return {'is_successful': True}

        except AttributeError:
            return response

    function_wrapper.__name__ = function.__name__
    function_wrapper.__doc__ = function.__doc__
    return function_wrapper


class APIError(Exception):

    def __init__(self, status, errors, url):
        self.status = status
        self.errors = errors
        self.url = url

    def __str__(self):
        return "HTTP Status Code {}: {} at URL {}".format(
            self.status,
            self.errors,
            self.url
        )



class AirMassModelTypeEnum(object):
    """Air Mass Model"""
    BIRD_HULSTROM = 0
    KASTEN_SANDIA = 1


class BacktrackingTypeEnum(object):
    """Backtracking Type"""
    TRUE_TRACKING = 0   # no backtracking
    BACKTRACKING = 1    # shade avoidance


class CellDesignTypeEnum(object):
    """Solar Cell Design Types"""
    FULL_CELL = 0
    HALF_CELL = 1
    NA = 2
    NONE_SPECIFIED = 3


class CellTechnologyTypeEnum(object):
    """Cell Technology"""
    NTYPE_MONO_CSI = 1
    PTYPE_MONO_CSI_PERC = 2
    PTYPE_MONO_CSI_BSF = 3
    POLY_CSI_PERC = 4
    POLY_CSI_BSF = 5
    CDTE = 6
    CIGS = 7


class CleaningFrequencyEnum(object):
    """Cleaning Frequency"""
    NONE = 0
    DAILY = 1
    MONTHLY = 2
    QUARTERLY = 3
    YEARLY = 4


class ConstructionTypeEnum(object):
    """Construction Type"""
    GLASS_GLASS = 1
    GLASS_BACKSHEET = 2


class DataSourceEnum(object):
    """Data Source"""
    MANUFACTURER = 1
    PVSYST = 2
    UNIVERSITY_OF_GENEVA = 3
    PHOTON = 4
    SANDIA_DATABASE = 5
    CUSTOM = 6


class DegradationModelEnum(object):
    """Degradation Model"""
    NONE = 0
    STEPPED_AC = 1
    LINEAR_AC = 2
    LINEAR_DC = 3
    NON_LINEAR_DC = 4


class DiffuseDirectDecompositionModelEnum(object):
    """Diffuse Direct Decomposition Model"""
    ERBS = 0
    REINDL = 1
    DIRINT = 2
    NONE = 3


class DiffuseShadingModelEnum(object):
    """Diffuse Shading Model"""
    NONE = 0
    SCHAAR_PANCHULA = 1


class DirectBeamShadingModelEnum(object):
    """Direct Beam Shading Model"""
    LINEAR = 0
    NONE = 1
    TWO_DIMENSION = 2       # Retired
    FRACTIONAL_EFFECT = 3   # Fractional Electric Shading
    CSI_3_DIODE = 4
    MODULE_FILE_DEFINED = 5


class EntityTypeEnum(object):
    """Entity Type"""
    PROJECT = 1
    MODULE = 2
    INVERTER = 3
    WEATHER = 4
    PREDICTION = 5


class ESSChargeAlgorithmEnum(object):
    """Energy Storage System (ESS) Charge Algorithm"""
    LGIA_EXCESS = 0
    ENERGY_AVAILABLE = 1
    CUSTOM = 2


class ESSDispatchCustomCommandEnum(object):
    """Energy Storage System (ESS) Dispatch Custom Command"""
    NONE = 0
    DISCHARGE = 1
    CHARGE = 2


class FacialityEnum(object):
    """Faciality"""
    MONOFACIAL = 0
    BIFACIAL = 1


class IncidenceAngleModelTypeEnum(object):
    """Incidence Angle Model Type"""
    SANDIA = 2
    ASHRAE = 3
    NONE = 4
    TABULAR_IAM = 5
    PHYSICAL = 6


class InverterTypeEnum(object):
    """Inverter Types"""
    UNSPECIFIED = 0
    STRING = 1
    CENTRAL = 2


class LibraryStatusEnum(object):
    """Library Status (for Module, Inverter, Weather)"""
    UNKNOWN = 0
    DRAFT_PRIVATE = 1
    DRAFT_SHARED = 2
    ACTIVE = 3
    RETIRED = 4
    GLOBAL = 5
    GLOBAL_RETIRED = 6


class ModuleDegradationModelEnum(object):
    """Module Degradation Model"""
    UNSPECIFIED = 0
    LINEAR = 1
    NONLINEAR = 2


class ModuleOrientationEnum(object):
    """Module Orientation"""
    LANDSCAPE = 0
    PORTRAIT = 1


class ModuleShadingResponseEnum(object):
    """Module Shading Response"""
    NONE = 0
    LINEAR = 1
    FRACTIONAL_EFFECT = 2   # Fractional Electrical Shading
    CSI_3_DIODE = 3
    CUSTOM = 4


class ModuleTemperatureModelEnum(object):
    """Module Temperature Model"""
    HEAT_BALANCE = 0
    SANDIA = 1
    NOCT = 2


class ModuleTypeEnum(object):
    """Module Type"""
    SINGLE_DIODE = 0
    ADVANCED_DIODE = 1


class PerezModelCoefficientsEnum(object):
    """Perez Coefficients"""
    PLANT_PREDICT = 0
    ALL_SITES_COMPOSITE_1990 = 1
    ALL_SITES_COMPOSITE_1988 = 2
    SANDIA_COMPOSITE_1988 = 3
    USA_COMPOSITE_1988 = 4
    FRANCE_1988 = 5
    PHOENIX_1988 = 6
    ELMONTE_1988 = 7
    OSAGE_1988 = 8
    ALBUQUERQUE_1988 = 9
    CAPE_CANAVERAL_1988 = 10
    ALBANY_1988 = 11


class PredictionStatusEnum(object):
    """Prediction Status"""
    DRAFT_PRIVATE = 1
    DRAFT_SHARED = 2
    ANALYSIS = 3
    BID = 4
    CONTRACT = 5
    DEVELOPMENT = 6
    AS_BUILT = 7
    WARRANTY = 8
    ARCHIVED = 9


class PredictionVersionEnum(object):
    """Prediction Version"""
    VERSION_3 = 3
    VERSION_4 = 4
    VERSION_5 = 5
    VERSION_6 = 6
    VERSION_7 = 7
    VERSION_8 = 8
    VERSION_9 = 9
    VERSION_10 = 10
    VERSION_11 = 11


class ProcessingStatusEnum(object):
    """Processing Status"""
    NONE = 0
    QUEUED = 1
    RUNNING = 2
    SUCCESS = 3
    ERROR = 4


class ProjectStatusEnum(object):
    """Project Status"""
    ACTIVE = 0
    ARCHIVED = 1


class PVModelTypeEnum(object):
    """PV Model"""
    ONE_DIODE_RECOMBINATION = 0
    ONE_DIODE = 1
    ONE_DIODE_RECOMBINATION_NONLINEAR = 3


class SoilingModelTypeEnum(object):
    """Soiling Model"""
    CONSTANT_MONTHLY = 0
    WEATHER_FILE = 1
    NONE = 2


class SpectralShiftModelEnum(object):
    """Spectral Shift Model"""
    NO_SPECTRAL_SHIFT = 0
    ONE_PARAM_PWAT_OR_SANDIA = 1
    TWO_PARAM_PWAT_AND_AM = 2
    MONTHLY_OVERRIDE = 3


class SpectralWeatherTypeEnum(object):
    """Spectral Weather Type"""
    NONE = 0
    NGAN_PWAT = 1
    NGAN_RH = 2
    NGAN_DEWPOINT = 3


class TrackingTypeEnum(object):
    """Tracking Type"""
    FIXED_TILT = 0
    HORIZONTAL_TRACKER = 1
    SEASONAL_TILT = 2


class TranspositionModelEnum(object):
    """Transposition Model"""
    HAY = 0
    PEREZ = 1

class CircumsolarTreatmentTypeEnum(object):
    "Circumsolar Allocation Type"
    DIFFUSE = 0
    DIRECT = 1


class WeatherDataProviderEnum(object):
    """Weather Data Provider"""
    NREL = 1
    AWS = 2
    WIND_LOGICS = 3
    METEONORM = 4
    THREE_TIER = 5
    CLEAN_POWER_RESEARCH = 6
    GEO_MODEL_SOLAR = 7
    GEO_SUN_AFRICA = 8
    SODA = 9
    HELIO_CLIM = 10
    SOLAR_RESOURCE_ASSESSMENT = 11
    ENERGY_PLUS = 12
    OTHER = 13
    CUSTOMER = 14
    SOLAR_PROSPECTOR = 15
    GLOBAL_FED = 16
    NSRDB = 17
    WHITE_BOX_TECHNOLOGIES = 18
    SOLARGIS = 19
    NASA = 20
    THREE_TIER_VAISALA=21
    SOLCAST=22



class WeatherDataTypeEnum(object):
    """Weather Data Type"""
    SYNTHETIC_MONTHLY = 0
    SATELLITE = 1
    GROUND_CORRECTED = 2
    MEASURED = 3
    TMY3 = 4
    TGY = 5
    TMY = 6
    PSM = 7
    SUNY = 8
    MTS2 = 9
    CZ2010 = 10


class WeatherFileColumnTypeEnum(object):
    """Weather File Column Type"""
    GHI = 1
    DNI = 2
    DHI = 3
    TEMP = 4
    WINDSPEED = 5
    RELATIVE_HUMIDITY = 6
    PWAT = 7
    RAIN = 8
    PRESSURE = 9
    DEWPOINT_TEMP = 10
    WIND_DIRECTION = 11
    SOILING_LOSS = 12
    POAI = 13
    REAR_POAI = 14


class WeatherPLevelEnum(object):
    """Weather P-Level"""
    P50 = 0
    P90 = 1
    P95 = 3
    P99 = 4
    NA = 2      # N/A
    P75 = 5


class WeatherSourceTypeAPIEnum(object):
    """
    Weather Source Type API (web-service downloadable vendors). This Enum is used when calling
    :py:meth:`~plantpredict.weather.Weather.download`.
    """
    UNKNOWN = 0   # Default
    METEONORM = 1
    CPR_SOLAR_ANYWHERE = 2
    NSRDB_PSM = 3
    NSRDB_SUNY = 4
    NSRDB_MTS2 = 5
    SOLAR_GIS = 6
    NASA = 7


class WeatherTimeResolution(object):
    """Weather Time Resolution"""
    UNKNOWN = 0
    HALF_HOUR = 1
    HOUR = 2
    MINUTE = 3

class TimeSeriesTypes(object):
    INVERTER_SET_POINT = 0
    INVERTER_DERATE = 1
    TRACKING_ANGLE = 2
    MODULE_SURFACE_TEMPERATURE = 3
    LGIA_LIMIT = 4
    IMPP_ADJUSTMENT = 5
    VMPP_ADJUSTMENT = 6



import json
import requests

from plantpredict.utilities import convert_json, camel_to_snake, decorate_all_methods
from plantpredict.error_handlers import handle_refused_connection, handle_error_response, APIError


@decorate_all_methods(handle_refused_connection)
@decorate_all_methods(handle_error_response)
class ASHRAE(object):
    """
    The :py:class:`~plantpredict.ashrae.ASHRAE` class is used to get key information for an ASHRAE station. It can be
    used on its own for any application, but mostly exists to find and assign plant design temperatures for a particular
    location to a :py:class:`~plantpredict.prediction.Prediction`.
    """
    def get_station(self, station_name=None):
        """
        Returns the ASHRAE station matching the specified name and shortest distance from the specified latitude and
        longitude. Sets the returned information as attributes on the instance of this class.

        :param str station_name: Valid name of ASHRAE weather station
        :return: # TODO once new http response is implemented
        """
        self.station_name = station_name if station_name else self.station_name
        response = requests.get(
            url=self.api.base_url + "/ASHRAE/GetStation",
            headers={"Authorization": "Bearer " + self.api.access_token},
            params={"latitude": self.latitude, "longitude": self.longitude, "stationName": self.station_name}
        )
        if not response.status_code == 200:
            raise APIError(response.status_code, response.content, response.url)

        attr = convert_json(response.json(), camel_to_snake)
        for key in attr:
            setattr(self, key, attr[key])

        return response

    def get_closest_station(self):
        """
        Returns the ASHRAE station with the shortest distance from the specified latitude and longitude. Sets the
        returned information as attributes on the instance of this class.

        :return: # TODO once new http response is implemented
        """
        response = requests.get(
            url=self.api.base_url + "/ASHRAE",
            headers={"Authorization": "Bearer " + self.api.access_token},
            params={"latitude": self.latitude, "longitude": self.longitude}
        )
        if not response.status_code == 200:
            raise APIError(response.status_code, response.content, response.url)

        attr = convert_json(response.json(), camel_to_snake)
        for key in attr:
            setattr(self, key, attr[key])

        return response

    def __init__(self, api, latitude=None, longitude=None, station_name=None):
        self.api = api
        self.latitude = latitude
        self.longitude = longitude
        self.station_name = station_name

        super(ASHRAE, self).__init__()



import requests
import json

from plantpredict.project import Project
from plantpredict.prediction import Prediction
from plantpredict.powerplant import PowerPlant
from plantpredict.geo import Geo
from plantpredict.inverter import Inverter
from plantpredict.module import Module
from plantpredict.weather import Weather
from plantpredict.ashrae import ASHRAE


class Api(object):

    def __get_access_token(self):
        """
        """
        response = requests.post(
            url=self.auth_url,
            headers={"content-type": "application/x-www-form-urlencoded"},
            params={
                "grant_type": "client_credentials",
                "scope": "transactions/post transactions/get",
                "client_id": self.client_id,
                "client_secret": self.client_secret
            }
        )

        # set authentication token as global variable
        try:
            self.access_token = response.json()['access_token']
        except KeyError:
            print("Authentification failed. Response:", response.text)
            pass

        return response

    def __init__(self, client_id, client_secret, base_url="https://api.plantpredict.terabase.energy",
                 auth_url="https://terabase-prd.auth.us-west-2.amazoncognito.com/oauth2/token"):
        self.base_url = base_url
        self.auth_url = auth_url

        self.client_id = client_id
        self.client_secret = client_secret

        self.access_token = None

        self.__get_access_token()

        super(Api, self).__init__()

    def project(self, **kwargs):
        return Project(self, **kwargs)

    def prediction(self, **kwargs):
        return Prediction(self, **kwargs)

    def powerplant(self, **kwargs):
        return PowerPlant(self, **kwargs)

    def geo(self, **kwargs):
        return Geo(self, **kwargs)

    def inverter(self, **kwargs):
        return Inverter(self, **kwargs)

    def module(self, **kwargs):
        return Module(self, **kwargs)

    def weather(self, **kwargs):
        return Weather(self, **kwargs)

    def ashrae(self, **kwargs):
        return ASHRAE(self, **kwargs)



import json
import requests
from plantpredict.plant_predict_entity import PlantPredictEntity
from plantpredict.error_handlers import handle_refused_connection, handle_error_response
from plantpredict.utilities import convert_json, camel_to_snake, snake_to_camel
from plantpredict.enumerations import EntityTypeEnum


class Weather(PlantPredictEntity):
    """
    The full contents of the Weather database entity (in JSON) can be found under
    "GET /Weather/{Id}" in `the general PlantPredict API documentation
    <https://api.plantpredict.terabase.energy/swagger/ui/index#!/Weather/Weather_Get_0>`_.
    """
    def create(self):
        """
        POST /Weather

        Creates a new Weather entity.

        .. container:: toggle

            .. container:: header

                **Required Attributes**

            .. container:: required_attributes

                .. csv-table:: Minimum required attributes for successful Weather creation
                    :delim: ;
                    :header: Field, Type, Description
                    :stub-columns: 1

                    name; str; Name of weather file
                    country_code; str; Country code of the Weather's location (ex. US for United States, AUS for Australia, etc.) :py:meth:`plantpredict.Geo.get_location_info` will return this information.
                    country; str; Full name of the country of the Weather's location. :py:meth:`plantpredict.Geo.get_location_info` will return this information.
                    latitude; float; North-South coordinate of the Weather location (in decimal degrees).
                    longitude; float; East-West coordinate of the Weather location (in decimal degrees).
                    data_provider; int; Represents a weather data source. See (and/or import) :py:mod:`plantpredict.enumerations.WeatherDataProviderEnum` for a list of options.
                    weather_details; list of dict; The code block below contains an example of one timestamp (array element) of this field, as well as information on which dictionary keys are required.

                .. code-block:: python

                    weather_details[109] = {
                        "index": 110,                           # REQUIRED | is equal to the list index + 1
                        "time_stamp": "2018-01-01T1:00:00",     # REQUIRED
                        "global_horizontal_irradiance": 139.3,  # REQUIRED if no 'plane_of_array_irradiance' | [W/m^2]
                        "diffuse_horizontal_irradiance": 139.3, # [W/m^2]
                        "direct_normal_irradiance": 0.0,        # [W/m^2]
                        "beam_horizontal_irradiance": 0.0,      # [W/m^2]
                        "plane_of_array_irradiance": 100.0,     # REQUIRED if no 'global_horizontal_irradiance' | [W/m^2]
                        "temperature": 1.94,                    # REQUIRED | [degrees-Celsius]
                        "relative_humidity": 74.5,              # [%]
                        "precipitable_water": 2.0,              # [cm]
                        "soiling_loss": 0.19                    # [%]
                    }

        :return: A dictionary containing the weather id.
        :rtype: dict
        """
        self.create_url_suffix = "/Weather"
        return super(Weather, self).create()

    def delete(self):
        """
        DELETE /Weather/{WeatherId}

        Deletes an existing Weather entity in PlantPredict. The local instance of the Weather entity must have
        attribute self.id identical to the weather id of the Weather to be deleted.

        :return: A dictionary {"is_successful": True}.
        :rtype: dict
        """
        self.delete_url_suffix = "/Weather/{}".format(self.id)

        return super(Weather, self).delete()

    def get(self):
        """
        GET /Weather/{Id}

        Retrieves an existing Weather entity in PlantPredict and automatically assigns all of its attributes to the
        local Weather object instance. The local instance of the Weather entity must have attribute self.id identical
        to the weather id of the Weather to be retrieved.

        :return: A dictionary containing all of the retrieved Weather attributes.
        :rtype: dict
        """
        self.get_url_suffix = "/Weather/{}".format(self.id)
        return super(Weather, self).get()

    def update(self):
        """
        PUT /Weather

        Updates an existing Weather entity in PlantPredict using the full attributes of the local Weather object instance.
        Calling this method is most commonly preceded by instantiating a local instance of Weather with a specified
        weather id, calling the Weather.get() method, and changing any attributes locally.

        The required fields are identical to those of :py:meth:`plantpredict.weather.create` with the addition of:
        .. csv-table:: Minimum required attributes for successful Weather creation
            :delim: ;
            :header: Field; Type; Description
            :stub-columns: 1

            id; int; Unique identifier of existing Weather entity.


        :return: A dictionary {"is_successful": True}.
        :rtype: dict
        """
        self.update_url_suffix = "/Weather/{}".format(self.id)
        return super(Weather, self).update()

    @handle_refused_connection
    @handle_error_response
    def get_details(self):
        """
        GET /Weather/{Id}/Detail

        Returns detailed time series of Weather entity.

        :returns: A list of dictionaries where each dictionary contains one timestamp of detailed weather data.
        :rtype: list of dicts
        """
        return requests.get(
            url=self.api.base_url + "/Weather/{}/Detail".format(self.id),
            headers={"Authorization": "Bearer " + self.api.access_token}
        )

    @handle_refused_connection
    @handle_error_response
    def search(self, latitude, longitude, search_radius=1):
        """
        GET /Weather/Search

        Searches for all existing Weather entities within a search radius of a specified latitude/longitude.

        :param latitude: North-South coordinate of the Weather location, in decimal degrees.
        :type latitude: float
        :param longitude: East-West coordinate of the Project location, in decimal degrees.
        :type longitude: float
        :param search_radius: search radius in miles
        :type search_radius: float
        :return: #TODO
        :rtype: list of dicts
        """

        response = requests.get(
            url=self.api.base_url + "/Weather/Search",
            headers={"Authorization": "Bearer " + self.api.access_token},
            params=convert_json({
                'latitude': latitude,
                'longitude': longitude,
                'search_radius': search_radius
            }, snake_to_camel)
        )

        weather_list = response.json()

        return [convert_json(w, camel_to_snake) for w in weather_list]

    @handle_refused_connection
    @handle_error_response
    def download(self, latitude, longitude, provider=0):
        """
        POST /Weather/Download/{Provider}

        :param latitude:
        :type latitude: float
        :param longitude:
        :type longitude: float
        :param provider: Represents a weather data source. See (and/or import) :py:mod:`plantpredict.enumerations.WeatherSourceTypeAPIEnum` for a list of options.
        :type provider: int
        :return: #TODO
        :rtype: dict
        """
        response = requests.post(
            url=self.api.base_url + "/Weather/Download/{}".format(provider),
            headers={"Authorization": "Bearer " + self.api.access_token},
            params={'latitude': latitude, 'longitude': longitude}
        )

        self.id = response.json()['id'] if 200 <= response.status_code < 300 else None

        return response

    @handle_refused_connection
    @handle_error_response
    def change_status(self, new_status, note=""):
        """
        POST /Weather/Status
        Change the status (and resulting sharing/privacy settings) of a weather file (ex. from py:attr:`DRAFT_PRIVATE` to
        py:attr:`DRAFT-SHARED`.
        :param int new_status: Enumeration representing status to change weather to. See (or import)
                               :py:class:`plantpredict.enumerations.LibraryStatusEnum`.
        :param str note: Description of reason for change.
        :return:
        """
        return requests.post(
            url=self.api.base_url + "/Weather/Status",
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=[{
                "name": self.name,
                "id": self.id,
                "type": EntityTypeEnum.WEATHER,
                "status": new_status,
                "note": note
            }]
        )

    @handle_refused_connection
    @handle_error_response
    def generate_weather(self):
        """
        Post /Weather/GenerateWeather

        Returns a synthetic weather time series based on monthly data. The monthly data must be defined as a list of dicts in a class attribute "monthly_values"

        :returns: A dictionary with all weather parameters, including and especially hourly synthetic data in "weather_details".
        :rtype: dict
        """
        return requests.post(
            url=self.api.base_url + "/Weather/GenerateWeather",
            headers={"Authorization": "Bearer " + self.api.access_token},
            json=convert_json(self.__dict__, snake_to_camel),
        )



import unittest
import os
import pandas as pd

from plantpredict.helpers import load_from_excel, export_to_excel


class TestHelpers(unittest.TestCase):
    def setUp(self):
        self.data = [
            {"Index": 1, "Name": "Stephen", "City": "Denver, CO"},
            {"Index": 2, "Name": "Kendra", "City": "Denver, CO"},
            {"Index": 3, "Name": "Sam", "City": "San Francisco, CA"},
            {"Index": 4, "Name": "Alex", "City": "Tempe, AZ"}
        ]

    def test_export_to_excel(self):
        export_to_excel(
            data=self.data,
            file_path="test_data/testing_helpers.xlsx",
            field_order=["Index", "Name", "City"],
            sorting_fields=["Index"]
        )

        self.assertTrue(os.path.exists("test_data/testing_helpers.xlsx"))

        xl = pd.ExcelFile("test_data/testing_helpers.xlsx")
        xl_truth = pd.ExcelFile("test_data/testing_helpers_truth.xlsx")
        self.assertEqual(
            xl.parse("Sheet1", index_col=None).to_dict('records'),
            xl_truth.parse("Sheet1", index_col=None).to_dict('records')
        )

        os.remove("test_data/testing_helpers.xlsx")

    def test_export_to_excel_with_sheet_name(self):
        export_to_excel(
            data=self.data,
            file_path="test_data/testing_helpers.xlsx",
            sheet_name="Sheet Name",
            field_order=["Index", "Name", "City"],
            sorting_fields=["Index"]
        )

        self.assertTrue(os.path.exists("test_data/testing_helpers.xlsx"))

        xl = pd.ExcelFile("test_data/testing_helpers.xlsx")
        xl_truth = pd.ExcelFile("test_data/testing_helpers_truth.xlsx")
        self.assertEqual(
            xl.parse("Sheet Name", index_col=None).to_dict('records'),
            xl_truth.parse("Sheet1", index_col=None).to_dict('records')
        )

        os.remove("test_data/testing_helpers.xlsx")

    def test_load_from_excel(self):
        loaded_data = load_from_excel("test_data/testing_helpers_truth.xlsx")
        self.assertEqual(loaded_data, self.data)

    def test_load_from_excel_with_sheet_name(self):
        loaded_data = load_from_excel("test_data/testing_helpers_truth.xlsx", "Sheet1")
        self.assertEqual(loaded_data, self.data)


if __name__ == '__main__':
    unittest.main()



import mock
import unittest
import json

from plantpredict.inverter import Inverter
from tests import plantpredict_unit_test_case, mocked_requests


class TestInverter(plantpredict_unit_test_case.PlantPredictUnitTestCase):
    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.create')
    def test_create(self, mocked_create):
        self._make_mocked_api()
        inverter = Inverter(api=self.mocked_api)

        inverter.create()
        self.assertEqual(inverter.create_url_suffix, "/Inverter")
        self.assertTrue(mocked_create.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.delete')
    def test_delete(self, mocked_delete):
        self._make_mocked_api()
        inverter = Inverter(api=self.mocked_api, id=808)

        inverter.delete()
        self.assertEqual(inverter.delete_url_suffix, "/Inverter/808")
        self.assertTrue(mocked_delete.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.get')
    def test_get(self, mocked_get):
        self._make_mocked_api()
        inverter = Inverter(api=self.mocked_api, id=808)

        inverter.get()
        self.assertEqual(inverter.get_url_suffix, "/Inverter/808")
        self.assertTrue(mocked_get.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.update')
    def test_update(self, mocked_update):
        self._make_mocked_api()
        inverter = Inverter(api=self.mocked_api, id=808)

        inverter.update()
        self.assertEqual(inverter.update_url_suffix, "/Inverter")
        self.assertTrue(mocked_update.called)

    @mock.patch('plantpredict.inverter.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_kva(self):
        self._make_mocked_api()
        inverter = Inverter(api=self.mocked_api, id=808)
        response = inverter.get_kva(elevation=1000, temperature=20, use_cooling_temp=True)

        self.assertEqual(response.json()['kva'], 700.0)


if __name__ == '__main__':
    unittest.main()



import mock
import unittest
import json

from plantpredict.module import Module
from tests import plantpredict_unit_test_case, mocked_requests


class TestModule(plantpredict_unit_test_case.PlantPredictUnitTestCase):
    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.create')
    def test_create(self, mocked_create):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)
        module.stc_short_circuit_current = 1.23
        module.stc_short_circuit_current_temp_coef = 0.04
        module.length = 2000
        module.width = 1200
        module.stc_max_power = 120.0

        module.create()
        self.assertEqual(module.create_url_suffix, "/Module")
        self.assertTrue(mocked_create.called)
        self.assertEqual(module.short_circuit_current_at_stc, 1.23)
        self.assertEqual(module.linear_temp_dependence_on_isc, 0.04)
        self.area = 2.4
        self.stc_efficiency = 120.0 / (2.4 * 1000.0)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.delete')
    def test_delete(self, mocked_delete):
        self._make_mocked_api()
        module = Module(api=self.mocked_api, id=808)

        module.delete()
        self.assertEqual(module.delete_url_suffix, "/Module/808")
        self.assertTrue(mocked_delete.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.get')
    def test_get(self, mocked_get):
        self._make_mocked_api()
        module = Module(api=self.mocked_api, id=808)

        module.get()
        self.assertEqual(module.get_url_suffix, "/Module/808")
        self.assertTrue(mocked_get.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.update')
    def test_update(self, mocked_update):
        self._make_mocked_api()
        module = Module(api=self.mocked_api, id=808)

        module.update()
        self.assertEqual(module.update_url_suffix, "/Module")
        self.assertTrue(mocked_update.called)

    def test_parse_full_iv_curves_template_no_sheet_name(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)
        iv_curves = module._parse_full_iv_curves_template("test_data/test_parse_full_iv_curves_template.xlsx")

        self.assertEqual(len(iv_curves), 2)
        self.assertEqual(len(iv_curves[0]["data_points"]), 200)
        self.assertEqual(iv_curves[0]["temperature"], 25)
        self.assertEqual(len(iv_curves[1]["data_points"]), 200)
        self.assertEqual(iv_curves[1]["temperature"], 50)
        self.assertAlmostEqual(iv_curves[0]["data_points"][3]["current"], 2.450329288)
        self.assertAlmostEqual(iv_curves[1]["data_points"][198]["voltage"], 224.9763771)

    def test_parse_full_iv_curves_template_with_sheet_name(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)
        iv_curves = module._parse_full_iv_curves_template(
            file_path="test_data/test_parse_full_iv_curves_template.xlsx",
            sheet_name="NewSheetName"
        )

        self.assertEqual(len(iv_curves), 1)
        self.assertEqual(len(iv_curves[0]["data_points"]), 200)
        self.assertEqual(iv_curves[0]["temperature"], 25)
        self.assertAlmostEqual(iv_curves[0]["data_points"][3]["current"], 2.450329288)

    def test_parse_key_iv_points_template_no_sheet_name(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)
        key_iv_points = module._parse_key_iv_points_template(file_path="test_data/test_parse_key_iv_points_template.xlsx")

        self.assertEqual(len(key_iv_points), 27)
        self.assertEqual(key_iv_points[5]["temperature"], 15)
        self.assertEqual(key_iv_points[5]["irradiance"], 1000)
        self.assertEqual(key_iv_points[5]["short_circuit_current"], 1.74346881517)
        self.assertEqual(key_iv_points[5]["mpp_voltage"], 74.21342493)

    def test_parse_key_iv_points_template_with_sheet_name(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)
        key_iv_points = module._parse_key_iv_points_template(
            file_path="test_data/test_parse_key_iv_points_template.xlsx",
            sheet_name="NewSheetName"
        )

        self.assertEqual(len(key_iv_points), 27)
        self.assertEqual(key_iv_points[5]["temperature"], 15)
        self.assertEqual(key_iv_points[5]["irradiance"], 1000)
        self.assertEqual(key_iv_points[5]["short_circuit_current"], 1.74346881517)
        self.assertEqual(key_iv_points[5]["mpp_voltage"], 74.21342493)

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_generate_iv_curve(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)
        iv_curve = module.generate_iv_curve()
        module.num_iv_points = 100

        self.assertEqual(iv_curve, [{"current": 1.2, "voltage": 100.0}])

    def test_process_iv_curves_no_inputs(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        with self.assertRaises(ValueError) as e:
            module.process_iv_curves()
        self.assertEqual(e.exception.args[0], "Either a file path to the .xslx template for Full IV Curves input or "
                                              "the properly formatted JSON-serializable data structure for Key IV "
                                              "Points input must be assigned as input. See the Python SDK "
                                              "documentation (https://plantpredict-python.readthedocs.io/en/latest/)"
                                              " for more information.")

    def test_process_iv_curves_conflicting_inputs(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        with self.assertRaises(ValueError) as e:
            module.process_iv_curves(
                file_path="fake_file_path.xlsx",
                iv_curve_data={
                    "temperature": 25, "irradiance": 1000, "data_points": [{"current": 1.25, "voltage": 20.0}]
                }
            )
        self.assertEqual(e.exception.args[0], "Only one input option may be specified.")

    @mock.patch('plantpredict.module.requests.post', mocked_requests.mocked_requests_post)
    def test_process_iv_curves_with_file(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        data = module.process_iv_curves(file_path="test_data/test_parse_full_iv_curves_template.xlsx")
        self.assertEqual(data, [
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            },
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            }
        ])

    @mock.patch('plantpredict.module.requests.post', mocked_requests.mocked_requests_post)
    def test_process_iv_curves_with_file(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        data = module.process_iv_curves(iv_curve_data=[
            {"temperature": 25, "irradiance": 1000, "data_points": [{"current": 1.25, "voltage": 20.0}]},
            {"temperature": 25, "irradiance": 1000, "data_points": [{"current": 1.25, "voltage": 20.0}]}
        ])
        self.assertEqual(data, [
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            },
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            }
        ])

    def test_process_key_iv_points_no_inputs(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        with self.assertRaises(ValueError) as e:
            module.process_key_iv_points()
        self.assertEqual(e.exception.args[0], "Either a file path to the .xslx template for Key IV Points input or the "
                                              "properly formatted JSON-serializable data structure for Key IV Points "
                                              "input must be assigned as input. See the Python SDK documentation "
                                              "(https://plantpredict-python.readthedocs.io/en/latest/) for more "
                                              "information.")

    def test_process_key_iv_points_conflicting_inputs(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        with self.assertRaises(ValueError) as e:
            module.process_key_iv_points(
                file_path="fake_file_path.xlsx",
                key_iv_points_data={
                    "temperature": 25,
                    "irradiance": 1000,
                    "short_circuit_current": 9.43,
                    "open_circuit_voltage": 46.39,
                    "mpp_current": 8.9598,
                    "mpp_voltage": 38.1285,
                    "max_power": 341.6237
                }
            )
        self.assertEqual(e.exception.args[0], "Only one input option may be specified.")

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_process_key_iv_points_with_file(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        response = module.process_key_iv_points(file_path="test_data/test_parse_key_iv_points_template.xlsx")
        self.assertEqual(response.json(), {
            "stc_short_circuit_current": 1.7592,
            "stc_open_circuit_voltage": 90.2189,
            "stc_mpp_current": 1.6084,
            "stc_mpp_voltage": 72.4938,
            "stc_short_circuit_current_temp_coef": 0.0519,
            "stc_open_circuit_voltage_temp_coef": -0.3081,
            "stc_power_temp_coef": -0.3535,
            "effective_irradiance_response": [
                {"temperature": 25, "irradiance": 1000, "relative_efficiency": 1.0},
                {"temperature": 25, "irradiance": 800, "relative_efficiency": 1.0039},
                {"temperature": 25, "irradiance": 600, "relative_efficiency": 1.0032},
                {"temperature": 25, "irradiance": 400, "relative_efficiency": 0.9925},
                {"temperature": 25, "irradiance": 200, "relative_efficiency": 0.9582},
            ]
        })
        self.assertEqual(module.stc_short_circuit_current, 1.7592)
        self.assertEqual(module.effective_irradiance_response, [
                {"temperature": 25, "irradiance": 1000, "relative_efficiency": 1.0},
                {"temperature": 25, "irradiance": 800, "relative_efficiency": 1.0039},
                {"temperature": 25, "irradiance": 600, "relative_efficiency": 1.0032},
                {"temperature": 25, "irradiance": 400, "relative_efficiency": 0.9925},
                {"temperature": 25, "irradiance": 200, "relative_efficiency": 0.9582},
        ])

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_process_key_iv_points_with_data(self):
        self._make_mocked_api()
        module = Module(api=self.mocked_api)

        response = module.process_key_iv_points(key_iv_points_data=[
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            },
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            }
        ])
        self.assertEqual(response.json(), {
            "stc_short_circuit_current": 1.7592,
            "stc_open_circuit_voltage": 90.2189,
            "stc_mpp_current": 1.6084,
            "stc_mpp_voltage": 72.4938,
            "stc_short_circuit_current_temp_coef": 0.0519,
            "stc_open_circuit_voltage_temp_coef": -0.3081,
            "stc_power_temp_coef": -0.3535,
            "effective_irradiance_response": [
                {"temperature": 25, "irradiance": 1000, "relative_efficiency": 1.0},
                {"temperature": 25, "irradiance": 800, "relative_efficiency": 1.0039},
                {"temperature": 25, "irradiance": 600, "relative_efficiency": 1.0032},
                {"temperature": 25, "irradiance": 400, "relative_efficiency": 0.9925},
                {"temperature": 25, "irradiance": 200, "relative_efficiency": 0.9582},
            ]
        })
        self.assertEqual(module.stc_short_circuit_current, 1.7592)
        self.assertEqual(module.effective_irradiance_response, [
                {"temperature": 25, "irradiance": 1000, "relative_efficiency": 1.0},
                {"temperature": 25, "irradiance": 800, "relative_efficiency": 1.0039},
                {"temperature": 25, "irradiance": 600, "relative_efficiency": 1.0032},
                {"temperature": 25, "irradiance": 400, "relative_efficiency": 0.9925},
                {"temperature": 25, "irradiance": 200, "relative_efficiency": 0.9582},
        ])

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_calculate_basic_data_at_conditions(self):
        self._make_mocked_api()
        module = Module(self.mocked_api)

        data = module.calculate_basic_data_at_conditions(temperature=25, irradiance=1000)
        self.assertEqual(data, [
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            },
            {
                "temperature": 25,
                "irradiance": 1000,
                "short_circuit_current": 9.43,
                "open_circuit_voltage": 46.39,
                "mpp_current": 8.9598,
                "mpp_voltage": 38.1285,
                "max_power": 341.6237
            }
        ])

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_calculate_effective_irradiance_response(self):
        self._make_mocked_api()
        module = Module(self.mocked_api)

        response = module.calculate_effective_irradiance_response()
        self.assertEqual(response.json(), [
            {'temperature': 25, 'irradiance': 1000, 'relative_efficiency': 1.0},
            {'temperature': 25, 'irradiance': 800, 'relative_efficiency': 1.02},
            {'temperature': 25, 'irradiance': 600, 'relative_efficiency': 1.001},
            {'temperature': 25, 'irradiance': 400, 'relative_efficiency': 0.99},
            {'temperature': 25, 'irradiance': 200, 'relative_efficiency': 0.97}
        ])

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_generate_single_diode_parameters_advanced(self):
        self._make_mocked_api()
        module = Module(self.mocked_api)

        response = module.generate_single_diode_parameters_advanced()
        self.assertEqual(response.json(), {
            "maximum_series_resistance": 6.0,
            "maximum_recombination_parameter": 2.5,
            "saturation_current_at_stc": 0.0000000012,
            "diode_ideality_factor_at_stc": 1.56,
            "linear_temp_dependence_on_gamma": -0.04,
            "light_generated_current": 1.8
        })
        self.assertEqual(module.diode_ideality_factor_at_stc, 1.56)

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_generate_single_diode_parameters_default(self):
        self._make_mocked_api()
        module = Module(self.mocked_api)

        response = module.generate_single_diode_parameters_default()
        self.assertEqual(response.json(), {
            "maximum_series_resistance": 6.0,
            "maximum_recombination_parameter": 2.5,
            "saturation_current_at_stc": 0.0000000012,
            "diode_ideality_factor_at_stc": 1.78,
            "linear_temp_dependence_on_gamma": -0.04,
            "light_generated_current": 1.8
        })
        self.assertEqual(module.diode_ideality_factor_at_stc, 1.78)

    @mock.patch('plantpredict.module.requests.post', new=mocked_requests.mocked_requests_post)
    def test_optimize_series_resistance(self):
        self._make_mocked_api()
        module = Module(self.mocked_api)

        response = module.optimize_series_resistance()
        self.assertEqual(response.json(), {
            "maximum_series_resistance": 6.0,
            "maximum_recombination_parameter": 2.5,
            "saturation_current_at_stc": 0.0000000012,
            "diode_ideality_factor_at_stc": 1.22,
            "linear_temp_dependence_on_gamma": -0.04,
            "light_generated_current": 1.8
        })
        self.assertEqual(module.diode_ideality_factor_at_stc, 1.22)


if __name__ == '__main__':
    unittest.main()



import unittest
import mock
import json

import plantpredict
from plantpredict.geo import Geo
from tests import plantpredict_unit_test_case, mocked_requests


class TestGeo(plantpredict_unit_test_case.PlantPredictUnitTestCase):
    @mock.patch('plantpredict.geo.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_location_info(self):
        self._make_mocked_api()
        geo = Geo(api=self.mocked_api, latitude=39.67, longitude=-105.21)
        response = geo.get_location_info()

        self.assertEqual(response.json(), {
            "country": "United States",
            "country_code": "US",
            "locality": "Morrison",
            "region": "North America",
            "state_province": "Colorado",
            "state_province_code": "CO"
        })
        self.assertEqual(geo.country, "United States")
        self.assertEqual(geo.country_code, "US")
        self.assertEqual(geo.locality, "Morrison")
        self.assertEqual(geo.region, "North America")
        self.assertEqual(geo.state_province, "Colorado")
        self.assertEqual(geo.state_province_code, "CO")

    @mock.patch('plantpredict.geo.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_elevation(self):
        self._make_mocked_api()
        geo = Geo(api=self.mocked_api, latitude=39.67, longitude=-105.21)
        response = geo.get_elevation()

        self.assertEqual(response.json(), {"elevation": 1965.96})
        self.assertEqual(geo.elevation, 1965.96)

    @mock.patch('plantpredict.geo.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_time_zone(self):
        self._make_mocked_api()
        geo = Geo(api=self.mocked_api, latitude=39.67, longitude=-105.21)
        response = geo.get_time_zone()

        self.assertEqual(response.json(), {"time_zone": -7.0})

    @mock.patch('plantpredict.api.requests.post', autospec=True)
    def test_init(self, mock_api_post):
        mock_api_post.return_value.ok = True
        mock_api_post.return_value.content = '''{"access_token":"dummy_access_token",
                                            "refresh_token":"dummy_refresh_token"}'''
        api = plantpredict.Api(client_id="0oakq", client_secret="IEdpr")

        geo = Geo(api=api, latitude=39.67, longitude=-105.21)
        self.assertIsInstance(geo, Geo)
        self.assertIsInstance(geo.api, plantpredict.Api)
        self.assertEqual(geo.latitude, 39.67)
        self.assertEqual(geo.longitude, -105.21)


if __name__ == '__main__':
    unittest.main()



import mock
import unittest

from tests import plantpredict_unit_test_case, mocked_requests
from tests.mocked_methods import mock_get_inverter_apparent_power, mock_get_inverter_kva_rating, \
    mock_calculate_default_post_height, mock_calculate_collector_bandwidth
from plantpredict.powerplant import PowerPlant
from plantpredict.enumerations import TrackingTypeEnum, ModuleOrientationEnum, BacktrackingTypeEnum


class TestPowerPlant(plantpredict_unit_test_case.PlantPredictUnitTestCase):

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.create')
    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_and_set_average_power_factor')
    def test_create(self, mock_calculate_and_set_average_power_factor, mocked_create):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)

        powerplant.create()
        self.assertEqual(powerplant.create_url_suffix, "/Project/7/Prediction/77/PowerPlant")
        self.assertTrue(mock_calculate_and_set_average_power_factor.called)
        self.assertTrue(mocked_create.called)
        self.assertEqual(powerplant.power_factor, 1.0)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.get')
    def test_get(self, mocked_get):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)

        powerplant.get()
        self.assertEqual(powerplant.get_url_suffix, "/Project/7/Prediction/77/PowerPlant")
        self.assertTrue(mocked_get.called)

    @mock.patch('plantpredict.plant_predict_entity.PlantPredictEntity.update')
    def test_update(self, mocked_update):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)

        powerplant.update()
        self.assertEqual(powerplant.update_url_suffix, "/Project/7/Prediction/77/PowerPlant")
        self.assertTrue(mocked_update.called)

    def _init_powerplant_structure(self):
        self.powerplant.blocks = [
            {"id": 1, "name": 1, "arrays": [
                {"id": 11, "name": 1, "inverters": [{"id": 111, "name": "A", "dc_fields": [{"id": 1111, "name": 1}]}]}
            ]}
        ]

    def test_calculate_sum_power_factors(self):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        powerplant.blocks = [{
            'repeater': 1,
            'arrays': [
                {
                    'repeater': 1,
                    'inverters': [
                        {'repeater': 5, 'power_factor': 1.0},
                        {'repeater': 1, 'power_factor': 0.9},
                        {'repeater': 3, 'power_factor': 0.8}
                    ]
                },
                {
                    'repeater': 1,
                    'inverters': [
                        {'repeater': 6, 'power_factor': 0.96},
                        {'repeater': 2, 'power_factor': 1.0}
                    ]
                }
            ],
        }]

        sum_power_factors = powerplant._calculate_sum_power_factors()
        self.assertAlmostEqual(sum_power_factors, 16.06)

    def test_calculate_num_inverters(self):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        powerplant.blocks = [
            {'arrays': [{'repeater': 1, 'inverters': [{'power_factor': 0.91703056769, 'repeater': 3}]}]},
            {'arrays': [{'repeater': 2, 'inverters': [{'power_factor': 1, 'repeater': 1}]}]},
            {'arrays': [{'repeater': 1, 'inverters': [{'power_factor': 0.95487627365, 'repeater': 1}]}]},
        ]

        num_inverters = powerplant._calculate_num_inverters()
        self.assertEqual(num_inverters, 6)

    def test_calculate_and_set_average_power_factor(self):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        powerplant.blocks = [
            {'arrays': [{'repeater': 1, 'inverters': [{'power_factor': 0.91703056769, 'repeater': 3}]}]},
            {'arrays': [{'repeater': 2, 'inverters': [{'power_factor': 1, 'repeater': 1}]}]},
            {'arrays': [{'repeater': 1, 'inverters': [{'power_factor': 0.95487627365, 'repeater': 1}]}]},
        ]

        powerplant._calculate_and_set_average_power_factor()
        self.assertAlmostEqual(powerplant.power_factor, 0.9509946627850558)

    def test_calculate_and_set_average_power_factor_zero_inverters(self):
        self._make_mocked_api()
        powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        powerplant.blocks = [{'arrays': [{'inverters': []}, {'inverters': []}]}]

        powerplant._calculate_and_set_average_power_factor()
        self.assertEqual(powerplant.power_factor, 0)

    def test_add_transformer_first(self):
        """Tests adding a transformer to a power plant with no existing transformers."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_transformer(
            rating=0.6,
            high_side_voltage=600,
            no_load_loss=1.1,
            full_load_loss=1.7,
            ordinal=1
        )
        self.assertEqual(self.powerplant.transformers, [{
            "rating": 0.6,
            "high_side_voltage": 600,
            "no_load_loss": 1.1,
            "full_load_loss": 1.7,
            "ordinal": 1
        }])

    def test_add_transformer_multiple(self):
        """Tests adding a transformer to a power plant with existing transformers."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_transformer(rating=0.6, high_side_voltage=600, no_load_loss=1.1,
                                        full_load_loss=1.7, ordinal=1)
        self.powerplant.add_transformer(rating=0.6, high_side_voltage=600, no_load_loss=1.1,
                                        full_load_loss=1.7, ordinal=1)

        self.assertEqual(self.powerplant.transformers, [
            {"rating": 0.6, "high_side_voltage": 600, "no_load_loss": 1.1, "full_load_loss": 1.7, "ordinal": 1},
            {"rating": 0.6, "high_side_voltage": 600, "no_load_loss": 1.1, "full_load_loss": 1.7, "ordinal": 1}
        ])

    def test_add_transmission_line_first(self):
        """Tests adding a transmission line to a powerplant with no existing transmission lines."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_transmission_line(
            length=3.0,
            resistance=0.1,
            number_of_conductors_per_phase=1,
            ordinal=1
        )
        self.assertEqual(self.powerplant.transmission_lines, [{
            "length": 3.0,
            "resistance": 0.1,
            "number_of_conductors_per_phase": 1,
            "ordinal": 1
        }])

    def test_add_transmission_line_multiple(self):
        """Tests adding a transmission line to a powerplant with existing transmission lines."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_transmission_line(length=3.0, resistance=0.1, number_of_conductors_per_phase=1, ordinal=1)
        self.powerplant.add_transmission_line(length=3.0, resistance=0.1, number_of_conductors_per_phase=1, ordinal=1)
        self.assertEqual(self.powerplant.transmission_lines, [
            {"length": 3.0, "resistance": 0.1, "number_of_conductors_per_phase": 1, "ordinal": 1},
            {"length": 3.0, "resistance": 0.1, "number_of_conductors_per_phase": 1, "ordinal": 1}
        ])

    def test_clone_block(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()
        cloned_block_name = self.powerplant.clone_block(block_id_to_clone=1)

        self.assertIsNotNone(cloned_block_name)
        self.assertEqual(cloned_block_name, 2)
        self.assertEqual(len(self.powerplant.blocks), 2)
        self.assertEqual(self.powerplant.blocks[-1], {"id": 1, "name": 2, "arrays": [
                {"id": 11, "name": 1, "inverters": [{"id": 111, "name": "A", "dc_fields": [{"id": 1111, "name": 1}]}]}
            ]})

    def test_add_block_first_default_inputs(self):
        """Tests adding a block to a power plant with no existing plots, and all default inputs."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        block_name = self.powerplant.add_block()

        self.assertEqual(len(self.powerplant.blocks), 1)
        self.assertEqual(block_name, 1)
        self.assertEqual(self.powerplant.blocks[0], {
            "name": 1,
            "use_energization_date": False,
            "energization_date": "",
            "arrays": []
        })

    def test_add_block_first_non_default_inputs(self):
        """Tests adding a block to a power plant with no existing plots, and non-default inputs."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        block_name = self.powerplant.add_block(use_energization_date=True, energization_date='2018-01-01')

        self.assertEqual(len(self.powerplant.blocks), 1)
        self.assertEqual(block_name, 1)
        self.assertEqual(self.powerplant.blocks[0], {
            "name": 1,
            "use_energization_date": True,
            "energization_date": "2018-01-01",
            "arrays": []
        })

    def test_add_block_multiple(self):
        """Tests adding a block to a power plant with existing plots."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()
        block_name = self.powerplant.add_block()

        self.assertEqual(len(self.powerplant.blocks), 2)
        self.assertEqual(block_name, 2)
        self.assertEqual(self.powerplant.blocks[-1], {
            "name": 2,
            "use_energization_date": False,
            "energization_date": "",
            "arrays": []
        })

    def test_add_array_default_inputs(self):
        """Test adding array with all default inputs."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()
        array_name = self.powerplant.add_array(block_name=1, description="testing")

        self.assertEqual(len(self.powerplant.blocks[0]["arrays"]), 2)
        self.assertEqual(array_name, 2)
        self.assertEqual(self.powerplant.blocks[0]["arrays"][1], {
            "name": 2,
            "repeater": 1,
            "ac_collection_loss": 1,
            "das_load": 800,
            "cooling_load": 0,
            "additional_losses": 0,
            "transformer_enabled": True,
            "match_total_inverter_kva": True,
            "transformer_high_side_voltage": 34.5,
            "transformer_no_load_loss": 0.2,
            "transformer_full_load_loss": 0.7,
            "inverters": [],
            "description": "testing"
        })

    def test_validate_block_name(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        with self.assertRaises(ValueError):
            self.powerplant._validate_block_name(2)

    def test_add_array_invalid_block_name(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        with self.assertRaises(ValueError):
            self.powerplant.add_array(block_name=2, description="testing")

    def test_add_array_non_default_inputs(self):
        """Test add array with all non default inputs, including False for match_total_inverter_kva."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()
        array_name = self.powerplant.add_array(
            block_name=1,
            transformer_enabled=False,
            match_total_inverter_kva=False,
            transformer_kva_rating=700.0,
            repeater=2,
            ac_collection_loss=2,
            das_load=900,
            cooling_load=0.1,
            additional_losses=0.2,
            transformer_high_side_voltage=35.0,
            transformer_no_load_loss=0.3,
            transformer_full_load_loss=0.8,
            description="testing"

        )

        self.assertEqual(len(self.powerplant.blocks[0]["arrays"]), 2)
        self.assertEqual(array_name, 2)
        self.assertEqual(self.powerplant.blocks[0]["arrays"][1], {
            "name": 2,
            "repeater": 2,
            "ac_collection_loss": 2,
            "das_load": 900,
            "cooling_load": 0.1,
            "additional_losses": 0.2,
            "transformer_enabled": False,
            "match_total_inverter_kva": False,
            "transformer_high_side_voltage": 35,
            "transformer_no_load_loss": 0.3,
            "transformer_full_load_loss": 0.8,
            "inverters": [],
            "description": "testing",
            "transformer_kva_rating": 700.0
        })

    def test_validate_array_name_invalid_block(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        try:
            self.powerplant._validate_array_name(block_name=2, array_name=3)
        except ValueError as e:
            self.assertEqual(e.args[0], "2 is not a valid block name in the existing power plant structure.")

    def test_validate_array_name_invalid_array(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        try:
            self.powerplant._validate_array_name(block_name=1, array_name=3)
        except ValueError as e:
            self.assertEqual(e.args[0], "3 is not a valid array name in block 1.")

    def test_validate_inverter_setpoint_inputs_setpoint_none(self):
        """Tests _validate_inverter_setpoint_inputs when setpoint_kw is None."""
        setpoint_kw, power_factor = PowerPlant._validate_inverter_setpoint_inputs(
            setpoint_kw=None,
            power_factor=1.2,
            kva_rating=600
        )
        self.assertEqual(setpoint_kw, 720)
        self.assertEqual(power_factor, 1.2)

    def test_validate_inverter_setpoint_inputs_setpoint_provided_power_factor_1(self):
        """Tests _validate_inverter_setpoint_inputs when setpoint_kw is None."""
        setpoint_kw, power_factor = PowerPlant._validate_inverter_setpoint_inputs(
            setpoint_kw=800,
            power_factor=1.0,
            kva_rating=1000
        )
        self.assertEqual(setpoint_kw, 800)
        self.assertEqual(power_factor, 0.8)

    def test_validate_inverter_setpoint_inputs_invalid_inputs(self):
        with self.assertRaises(ValueError):
            PowerPlant._validate_inverter_setpoint_inputs(
                setpoint_kw=800,
                power_factor=1.2,
                kva_rating=1000
            )

    @mock.patch('plantpredict.powerplant.PowerPlant._get_inverter_apparent_power', mock_get_inverter_apparent_power)
    @mock.patch('plantpredict.powerplant.PowerPlant._get_inverter_kva_rating', mock_get_inverter_kva_rating)
    def test_add_inverter_default_inputs_use_cooling_temp(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self.powerplant.use_cooling_temp = True
        self._init_powerplant_structure()

        inverter_name = self.powerplant.add_inverter(block_name=1, array_name=1, inverter_id=123)
        self.assertEqual(len(self.powerplant.blocks[0]["arrays"][0]["inverters"]), 2)
        self.assertEqual(inverter_name, "B")
        self.assertEqual(self.powerplant.blocks[0]["arrays"][0]["inverters"][1], {
            "name": "B",
            "repeater": 1,
            "inverter_id": 123,
            "setpoint_kw": 900,
            "power_factor": 1.0,
            "kva_rating": 900.0,
            "dc_fields": []
        })

    @mock.patch('plantpredict.powerplant.PowerPlant._get_inverter_apparent_power', mock_get_inverter_apparent_power)
    @mock.patch('plantpredict.powerplant.PowerPlant._get_inverter_kva_rating', mock_get_inverter_kva_rating)
    def test_add_inverter_invalid_array_name(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self.powerplant.use_cooling_temp = True
        self._init_powerplant_structure()

        with self.assertRaises(ValueError):
            self.powerplant.add_inverter(block_name=1, array_name=3, inverter_id=123)

    @mock.patch('plantpredict.powerplant.PowerPlant._get_inverter_apparent_power', mock_get_inverter_apparent_power)
    def test_add_inverter_non_default_inputs_no_use_cooling_temp(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self.powerplant.use_cooling_temp = False
        self._init_powerplant_structure()
        inverter_name = self.powerplant.add_inverter(
            block_name=1, array_name=1, inverter_id=123, setpoint_kw=800
        )

        self.assertEqual(len(self.powerplant.blocks[0]["arrays"][0]["inverters"]), 2)
        self.assertEqual(inverter_name, "B")
        self.assertEqual(self.powerplant.blocks[0]["arrays"][0]["inverters"][1], {
            "name": "B",
            "repeater": 1,
            "inverter_id": 123,
            "setpoint_kw": 800,
            "power_factor": 1.0,
            "kva_rating": 800.0,
            "dc_fields": []
        })

    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_get_default_module_azimuth_from_latitude_above_equator(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)

        default_module_azimuth = self.powerplant._get_default_module_azimuth_from_latitude()
        self.assertEqual(default_module_azimuth, 180.0)

    def test_calculate_collector_bandwidth_portrait(self):
        collector_bandwidth = PowerPlant._calculate_collector_bandwidth(
            module_length=1960,
            module_width=992,
            module_orientation=ModuleOrientationEnum.PORTRAIT,
            modules_high=4,
            vertical_intermodule_gap=0.02
        )
        self.assertAlmostEqual(collector_bandwidth, 7.90, 2)

    def test_calculate_collector_bandwidth_landscape(self):
        collector_bandwidth = PowerPlant._calculate_collector_bandwidth(
            module_length=1960,
            module_width=992,
            module_orientation=ModuleOrientationEnum.LANDSCAPE,
            modules_high=4,
            vertical_intermodule_gap=0.02
        )
        self.assertAlmostEqual(collector_bandwidth, 4.03, 2)

    def test_calculate_table_length_portrait(self):
        table_length = PowerPlant._calculate_table_length(
            modules_wide=18,
            module_orientation=ModuleOrientationEnum.PORTRAIT,
            module_length=1960,
            module_width=992,
            lateral_intermodule_gap=0.02
        )
        self.assertEqual(table_length, 18.196)

    def test_calculate_table_length_landscape(self):
        table_length = PowerPlant._calculate_table_length(
            modules_wide=18,
            module_orientation=ModuleOrientationEnum.LANDSCAPE,
            module_length=1960,
            module_width=992,
            lateral_intermodule_gap=0.02
        )
        self.assertAlmostEqual(table_length, 35.62)

    def test_calculate_tables_per_row(self):
        tables_per_row = PowerPlant._calculate_tables_per_row(
            field_dc_power=756,
            planned_module_rating=360,
            modules_high=4,
            modules_wide=18,
            number_of_rows=10
        )
        self.assertEqual(tables_per_row, 3)

    def test_calculate_tables_per_row_with_tables_removed(self):
        tables_per_row = PowerPlant._calculate_tables_per_row(
            field_dc_power=756,
            planned_module_rating=360,
            modules_high=4,
            modules_wide=18,
            number_of_rows=10,
            tables_removed_for_pcs=1.0,

        )
        self.assertEqual(tables_per_row, 4)

    def test_calculate_dc_field_size_by_collector_bandwidth(self):
        dc_field_size = PowerPlant._calculate_dc_field_size_by_collector_bandwidth(
            number_of_rows=10,
            post_to_post_spacing=9.799999999999997,
            collector_bandwidth=4.03
        )
        self.assertAlmostEqual(dc_field_size, 92.23, 2)

    def test_calculate_dc_field_size_by_tables_per_row_landscape_module_orientation(self):
        dc_field_size = PowerPlant._calculate_dc_field_size_by_tables_per_row(
            tables_per_row=2.8463,
            module_orientation=ModuleOrientationEnum.LANDSCAPE,
            module_length=1960,
            module_width=992,
            lateral_intermodule_gap=0.02,
            modules_wide=18
        )
        self.assertAlmostEqual(dc_field_size, 101.42, 2)

    def test_calculate_dc_field_size_by_tables_per_row_portrait_module_orientation(self):
        dc_field_size = PowerPlant._calculate_dc_field_size_by_tables_per_row(
            tables_per_row=4,
            module_orientation=ModuleOrientationEnum.PORTRAIT,
            module_length=1960,
            module_width=992,
            lateral_intermodule_gap=0.02,
            modules_wide=18
        )
        self.assertAlmostEqual(dc_field_size, 72.86, 1)

    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_dc_field_size_by_tables_per_row')
    def test_calculate_dc_field_length_tracker(self, mock_calculate_dc_field_size_by_tables_per_row):
        self._make_mocked_api()
        powerplant = PowerPlant(self.mocked_api)
        powerplant._calculate_dc_field_length(2, ModuleOrientationEnum.PORTRAIT, 1962, 900, 0.02, 18,
                                              TrackingTypeEnum.HORIZONTAL_TRACKER, 1, 1.5, 4.03)
        self.assertTrue(mock_calculate_dc_field_size_by_tables_per_row.called)

    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_dc_field_size_by_collector_bandwidth')
    def test_calculate_dc_field_length_fixed_tilt(self, mock_calculate_dc_field_size_by_collector_bandwidth):
        self._make_mocked_api()
        powerplant = PowerPlant(self.mocked_api)
        powerplant._calculate_dc_field_length(2, ModuleOrientationEnum.PORTRAIT, 1962, 900, 0.02, 18,
                                              TrackingTypeEnum.FIXED_TILT, 1, 1.5, 4.03)
        self.assertTrue(mock_calculate_dc_field_size_by_collector_bandwidth.called)

    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_dc_field_size_by_collector_bandwidth')
    def test_calculate_dc_field_width_tracker(self, mock_calculate_dc_field_size_by_collector_bandwidth):
        self._make_mocked_api()
        powerplant = PowerPlant(self.mocked_api)
        powerplant._calculate_dc_field_width(TrackingTypeEnum.HORIZONTAL_TRACKER, 1, 1.5, 4.03, 2,
                                             ModuleOrientationEnum.PORTRAIT, 1962, 900, 0.02, 18)
        self.assertTrue(mock_calculate_dc_field_size_by_collector_bandwidth.called)

    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_dc_field_size_by_tables_per_row')
    def test_calculate_dc_field_width_fixed_tilt(self, mock_calculate_dc_field_size_by_tables_per_row):
        self._make_mocked_api()
        powerplant = PowerPlant(self.mocked_api)
        powerplant._calculate_dc_field_width(TrackingTypeEnum.FIXED_TILT, 1, 1.5, 4.03, 2,
                                             ModuleOrientationEnum.PORTRAIT, 1962, 900, 0.02, 18)
        self.assertTrue(mock_calculate_dc_field_size_by_tables_per_row.called)

    def test_validate_dc_field_sizing_both_specified(self):
        try:
            PowerPlant._validate_dc_field_sizing(
                field_dc_power=869.4,
                number_of_series_strings_wired_in_parallel=420.0,
                modules_wired_in_series=6,
                planned_module_rating=420.0
            )
        except ValueError as e:
            self.assertEqual(e.args[0], "Both field_dc_power and number_of_series_strings_wired_in_parallel are not "
                                        "None. Only one of these values can be specified (and the other will be "
                                        "calculated).")

    def test_validate_dc_field_sizing_with_field_dc_power(self):
        field_dc_power, number_of_series_strings_wired_in_parallel = PowerPlant._validate_dc_field_sizing(
            field_dc_power=869.4,
            number_of_series_strings_wired_in_parallel=None,
            modules_wired_in_series=6,
            planned_module_rating=420.0
        )
        self.assertEqual(field_dc_power, 869.4)
        self.assertEqual(number_of_series_strings_wired_in_parallel, 345.0)

    def test_validate_dc_field_sizing_with_number_of_series_strings_wired_in_parallel(self):
        field_dc_power, number_of_series_strings_wired_in_parallel = PowerPlant._validate_dc_field_sizing(
            field_dc_power=None,
            number_of_series_strings_wired_in_parallel=345.0,
            modules_wired_in_series=6,
            planned_module_rating=420.0
        )
        self.assertEqual(field_dc_power, 869.4)
        self.assertEqual(number_of_series_strings_wired_in_parallel, 345.0)

    def test_validate_dc_field_sizing_neither_specified(self):
        try:
            PowerPlant._validate_dc_field_sizing(
                field_dc_power=None,
                number_of_series_strings_wired_in_parallel=None,
                modules_wired_in_series=6,
                planned_module_rating=420.0
            )
        except ValueError as e:
            self.assertEqual(e.args[0], "Both field_dc_power and number_of_series_strings_wired_in_parallel are None. "
                                        "One of these variables must be specified, and the other will be calculated.")

    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_collector_bandwidth',
                new=mock_calculate_collector_bandwidth)
    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    def test_calculate_post_to_post_spacing_from_gcr(self):
        self._make_mocked_api()
        powerplant = PowerPlant(self.mocked_api)
        post_to_post_spacing = powerplant.calculate_post_to_post_spacing_from_gcr(
            ground_coverage_ratio=0.40,
            module_id=123,
            modules_high=4
        )
        self.assertAlmostEqual(post_to_post_spacing, 4.175, 3)

    def test_calculate_field_dc_power(self):
        field_dc_power = PowerPlant.calculate_field_dc_power_from_dc_ac_ratio(
            dc_ac_ratio=1.20,
            inverter_setpoint=630
        )
        self.assertEqual(field_dc_power, 756)

    def test_validate_mounting_structure_parameters_fixed_tilt_valid(self):
        PowerPlant._validate_mounting_structure_parameters(
            tracking_type=TrackingTypeEnum.FIXED_TILT,
            module_tilt=30.0,
            tracking_backtracking_type=None
        )

    def test_validate_mounting_structure_parameters_fixed_tilt_invalid(self):
        with self.assertRaises(ValueError) as e:
            PowerPlant._validate_mounting_structure_parameters(
                tracking_type=TrackingTypeEnum.FIXED_TILT,
                module_tilt=None,
                tracking_backtracking_type=None
            )

        self.assertEqual(e.exception.args[0], "The input module_tilt is required for a fixed tilt DC field.")

    def test_validate_mounting_structure_parameters_tracking_true_tracking_valid(self):
        PowerPlant._validate_mounting_structure_parameters(
            tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
            module_tilt=None,
            tracking_backtracking_type=BacktrackingTypeEnum.TRUE_TRACKING
        )

    def test_validate_mounting_structure_parameters_tracking_backtracking_valid(self):
        PowerPlant._validate_mounting_structure_parameters(
            tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
            module_tilt=None,
            tracking_backtracking_type=BacktrackingTypeEnum.BACKTRACKING
        )

    def test_validate_mounting_structure_parameters_tracking_invalid(self):
        with self.assertRaises(ValueError) as e:
            PowerPlant._validate_mounting_structure_parameters(
                tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
                module_tilt=None,
                tracking_backtracking_type=None
            )

        self.assertEqual(e.exception.args[0], "The input tracking_backtracking_type is required for a horizontal "
                                              "tracker DC field.")

    def test_validate_inverter_name_invalid_block(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        try:
            self.powerplant._validate_inverter_name(block_name=2, array_name=3, inverter_name='B')
        except ValueError as e:
            self.assertEqual(e.args[0], "2 is not a valid block name in the existing power plant structure.")

    def test_validate_inverter_name_invalid_array(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        try:
            self.powerplant._validate_inverter_name(block_name=1, array_name=3, inverter_name='B')
        except ValueError as e:
            self.assertEqual(e.args[0], "3 is not a valid array name in block 1.")

    def test_validate_inverter_name_invalid_inverter_name(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        try:
            self.powerplant._validate_inverter_name(block_name=1, array_name=1, inverter_name='B')
        except ValueError as e:
            self.assertEqual(e.args[0], "'B' is not a valid inverter name in array 1 of block 1.")

    def test_calculate_default_post_height_less_than_1pt5meters(self):
        """Tests that 1.5 is returned if calculated post height is less than 1.5"""
        post_height = PowerPlant._calculate_default_post_height(
            tracking_type=TrackingTypeEnum.FIXED_TILT,
            collector_bandwidth=1.5,
            module_tilt=30,
            minimum_tracking_limit_angle_d=-60,
            maximum_tracking_limit_angle_d=60
        )

        self.assertEqual(post_height, 1.5)

    def test_calculate_default_post_height_fixed_tilt(self):
        post_height = PowerPlant._calculate_default_post_height(
            tracking_type=TrackingTypeEnum.FIXED_TILT,
            collector_bandwidth=3.0,
            module_tilt=30,
            minimum_tracking_limit_angle_d=-60,
            maximum_tracking_limit_angle_d=60
        )

        self.assertEqual(post_height, 1.75)

    def test_calculate_default_post_height_tracker(self):
        post_height = PowerPlant._calculate_default_post_height(
            tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
            collector_bandwidth=3.0,
            module_tilt=30,
            minimum_tracking_limit_angle_d=-60,
            maximum_tracking_limit_angle_d=60
        )

        self.assertEqual(post_height, 2.299038105676658)

    def test_calculate_modules_wide(self):
        modules_wide = PowerPlant._calculate_modules_wide(strings_wide=3, modules_wired_in_series=6)

        self.assertEqual(modules_wide, 18)

    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_default_post_height', mock_calculate_default_post_height)
    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_add_dc_field_with_bifacial_default_inputs(self):
        self._make_mocked_api(module_id=456)
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_dc_field(
            block_name=1,
            array_name=1,
            inverter_name="A",
            module_id=456,
            tracking_type=TrackingTypeEnum.FIXED_TILT,
            number_of_series_strings_wired_in_parallel=400,
            modules_high=4,
            modules_wired_in_series=10,
            post_to_post_spacing=1.0,
            module_tilt=30,
        )

        self.assertEqual(self.powerplant.blocks[0]["arrays"][0]["inverters"][0]["dc_fields"][1], {
            "name": 2,
            "module_id": 456,
            "tracking_type": TrackingTypeEnum.FIXED_TILT,
            "module_tilt": 30,
            "minimum_tracking_limit_angle_d": -60.0,
            "maximum_tracking_limit_angle_d": 60.0,
            "module_orientation": ModuleOrientationEnum.LANDSCAPE,
            "modules_high": 4,
            "module_azimuth": 180.0,
            "collector_bandwidth": 4.859999999999999,
            "post_to_post_spacing": 1.0,
            "planned_module_rating": 120,
            "modules_wired_in_series": 10,
            "field_dc_power": 480.0,
            "number_of_series_strings_wired_in_parallel": 400,
            "module_count": 4000.0,
            "module_quality": 1.0,
            "module_mismatch_coefficient": 1.0,
            "light_induced_degradation": 1.0,
            "dc_wiring_loss_at_stc": 1.5,
            "dc_health": 1.0,
            "heat_balance_conductive_coef": -3.47,
            "heat_balance_convective_coef": -0.0594,
            "sandia_conductive_coef": 30.7,
            "cell_to_module_temp_diff": 3.0,
            "sandia_convective_coef": 0.0,
            "tracker_load_loss": 0.0,
            "lateral_intermodule_gap": 0.02,
            "vertical_intermodule_gap": 0.02,
            "modules_wide": 10,
            "table_to_table_spacing": 0.0,
            "number_of_rows": 1,
            "table_length": 20.18,
            "tables_per_row": 100,
            'field_length': 4.859999999999999,
            'field_width': 2019.98,
            'post_height': 2.234,
            'structure_shading': 0.0,
            'backside_mismatch': 3.0
        })

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_add_dc_field_with_bifacial_non_default_inputs(self):
        self._make_mocked_api(module_id=456)
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_dc_field(
            block_name=1,
            array_name=1,
            inverter_name="A",
            module_id=456,
            tracking_type=TrackingTypeEnum.FIXED_TILT,
            number_of_series_strings_wired_in_parallel=400,
            modules_high=4,
            modules_wired_in_series=10,
            post_to_post_spacing=1.0,
            module_tilt=30,
            post_height=1.5,
            structure_shading=1.0,
            backside_mismatch=2.0
        )

        self.assertEqual(self.powerplant.blocks[0]["arrays"][0]["inverters"][0]["dc_fields"][1], {
            "name": 2,
            "module_id": 456,
            "tracking_type": TrackingTypeEnum.FIXED_TILT,
            "module_tilt": 30,
            "minimum_tracking_limit_angle_d": -60.0,
            "maximum_tracking_limit_angle_d": 60.0,
            "module_orientation": ModuleOrientationEnum.LANDSCAPE,
            "modules_high": 4,
            "module_azimuth": 180.0,
            "collector_bandwidth": 4.859999999999999,
            "post_to_post_spacing": 1.0,
            "planned_module_rating": 120,
            "modules_wired_in_series": 10,
            "field_dc_power": 480.0,
            "number_of_series_strings_wired_in_parallel": 400,
            "module_count": 4000.0,
            "module_quality": 1.0,
            "module_mismatch_coefficient": 1.0,
            "light_induced_degradation": 1.0,
            "dc_wiring_loss_at_stc": 1.5,
            "dc_health": 1.0,
            "heat_balance_conductive_coef": -3.47,
            "heat_balance_convective_coef": -0.0594,
            "sandia_conductive_coef": 30.7,
            "cell_to_module_temp_diff": 3.0,
            "sandia_convective_coef": 0.0,
            "tracker_load_loss": 0.0,
            "lateral_intermodule_gap": 0.02,
            "vertical_intermodule_gap": 0.02,
            "modules_wide": 10,
            "table_to_table_spacing": 0.0,
            "number_of_rows": 1,
            "table_length": 20.18,
            "tables_per_row": 100.0,
            'field_length': 4.859999999999999,
            'field_width': 2019.98,
            'post_height': 1.5,
            'structure_shading': 1.0,
            'backside_mismatch': 2.0
        })

    def test_add_dc_field_seasonal_tilt(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        with self.assertRaises(ValueError):
            self.powerplant.add_dc_field(
                block_name=1,
                array_name=1,
                inverter_name="A",
                module_id=123,
                tracking_type=TrackingTypeEnum.SEASONAL_TILT,
                number_of_series_strings_wired_in_parallel=400,
                modules_high=4,
                modules_wired_in_series=10,
                post_to_post_spacing=1.0,
                module_tilt=30
            )

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_add_dc_field_fixed_tilt(self):
        """Test minimum inputs for successfully adding fixed tilt DC field."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        dc_field_name = self.powerplant.add_dc_field(
            block_name=1,
            array_name=1,
            inverter_name="A",
            module_id=123,
            tracking_type=TrackingTypeEnum.FIXED_TILT,
            number_of_series_strings_wired_in_parallel=400,
            modules_high=4,
            modules_wired_in_series=10,
            post_to_post_spacing=1.0,
            module_tilt=30
        )

        self.assertEqual(len(self.powerplant.blocks[0]["arrays"][0]["inverters"][0]["dc_fields"][0]), 2)
        self.assertEqual(dc_field_name, 2)
        self.assertEqual(self.powerplant.blocks[0]["arrays"][0]["inverters"][0]["dc_fields"][1], {
            "name": 2,
            "module_id": 123,
            "tracking_type": TrackingTypeEnum.FIXED_TILT,
            "module_tilt": 30,
            "minimum_tracking_limit_angle_d": -60.0,
            "maximum_tracking_limit_angle_d": 60.0,
            "module_orientation": ModuleOrientationEnum.LANDSCAPE,
            "modules_high": 4,
            "module_azimuth": 180.0,
            "collector_bandwidth": 4.859999999999999,
            "post_to_post_spacing": 1.0,
            "planned_module_rating": 120,
            "modules_wired_in_series": 10,
            "field_dc_power": 480.0,
            "number_of_series_strings_wired_in_parallel": 400,
            "module_count": 4000.0,
            "module_quality": 1.0,
            "module_mismatch_coefficient": 1.0,
            "light_induced_degradation": 1.0,
            "dc_wiring_loss_at_stc": 1.5,
            "dc_health": 1.0,
            "heat_balance_conductive_coef": -3.47,
            "heat_balance_convective_coef": -0.0594,
            "sandia_conductive_coef": 30.7,
            "cell_to_module_temp_diff": 3.0,
            "sandia_convective_coef": 0.0,
            "tracker_load_loss": 0.0,
            "lateral_intermodule_gap": 0.02,
            "vertical_intermodule_gap": 0.02,
            "modules_wide": 10,
            "table_to_table_spacing": 0.0,
            "number_of_rows": 1,
            "table_length": 20.18,
            "tables_per_row": 100.0,
            'field_length': 4.859999999999999,
            'field_width': 2019.98,
            'post_height': 2.215,
        })

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_add_dc_field_tracking(self):
        """Test minimum inputs for successfully adding tracker DC field."""
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        dc_field_name = self.powerplant.add_dc_field(
            block_name=1,
            array_name=1,
            inverter_name="A",
            module_id=123,
            tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
            number_of_series_strings_wired_in_parallel=400,
            modules_high=4,
            modules_wired_in_series=10,
            post_to_post_spacing=1.0,
            tracking_backtracking_type=BacktrackingTypeEnum.BACKTRACKING
        )

        self.assertEqual(len(self.powerplant.blocks[0]["arrays"][0]["inverters"][0]["dc_fields"][0]), 2)
        self.assertEqual(dc_field_name, 2)
        self.assertEqual(self.powerplant.blocks[0]["arrays"][0]["inverters"][0]["dc_fields"][1], {
            "name": 2,
            "module_id": 123,
            "tracking_type": TrackingTypeEnum.HORIZONTAL_TRACKER,
            "tracking_backtracking_type": BacktrackingTypeEnum.BACKTRACKING,
            "minimum_tracking_limit_angle_d": -60.0,
            "maximum_tracking_limit_angle_d": 60.0,
            "module_orientation": ModuleOrientationEnum.LANDSCAPE,
            "modules_high": 4,
            "module_azimuth": 180.0,
            "collector_bandwidth": 4.859999999999999,
            "post_to_post_spacing": 1.0,
            "planned_module_rating": 120,
            "modules_wired_in_series": 10,
            "field_dc_power": 480.0,
            "number_of_series_strings_wired_in_parallel": 400,
            "module_count": 4000.0,
            "module_quality": 1.0,
            "module_mismatch_coefficient": 1.0,
            "light_induced_degradation": 1.0,
            "dc_wiring_loss_at_stc": 1.5,
            "dc_health": 1.0,
            "heat_balance_conductive_coef": -3.47,
            "heat_balance_convective_coef": -0.0594,
            "sandia_conductive_coef": 30.7,
            "cell_to_module_temp_diff": 3.0,
            "sandia_convective_coef": 0.0,
            "tracker_load_loss": 0.0,
            "lateral_intermodule_gap": 0.02,
            "vertical_intermodule_gap": 0.02,
            "modules_wide": 10,
            "table_to_table_spacing": 0.0,
            "number_of_rows": 1,
            "table_length": 20.18,
            "tables_per_row": 100,
            'field_length': 2019.98,
            'field_width': 4.859999999999999,
            'post_height': 3.1044417311961854,
        })

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.powerplant.PowerPlant._validate_dc_field_sizing')
    @mock.patch('plantpredict.powerplant.PowerPlant._validate_mounting_structure_parameters')
    @mock.patch('plantpredict.powerplant.PowerPlant._validate_inverter_name')
    def test_add_dc_field_validation_methods_called(self, mock_validate_inverter_name,
                                                    mock_validate_mounting_structure_parameters,
                                                    mock_validate_dc_field_sizing):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        try:
            self.powerplant.add_dc_field(
                block_name=1,
                array_name=1,
                inverter_name="A",
                module_id=123,
                tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
                number_of_series_strings_wired_in_parallel=400,
                modules_high=4,
                modules_wired_in_series=10,
                post_to_post_spacing=1.0,
                tracking_backtracking_type=BacktrackingTypeEnum.BACKTRACKING
            )
        except ValueError:
            pass

        self.assertTrue(mock_validate_inverter_name.called)
        self.assertTrue(mock_validate_mounting_structure_parameters.called)
        self.assertTrue(mock_validate_dc_field_sizing.called)

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_tables_per_row')
    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_table_length')
    @mock.patch('plantpredict.powerplant.PowerPlant._get_default_module_azimuth_from_latitude')
    def test_add_dc_field_helper_methods_called(self, mock_get_default_module_azimuth_from_latitude,
                                                mock_calculate_table_length, mock_calculate_tables_per_row):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_dc_field(
            block_name=1,
            array_name=1,
            inverter_name="A",
            module_id=123,
            tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
            number_of_series_strings_wired_in_parallel=400,
            modules_high=4,
            modules_wired_in_series=10,
            post_to_post_spacing=1.0,
            tracking_backtracking_type=BacktrackingTypeEnum.BACKTRACKING,
        )

        self.assertTrue(mock_get_default_module_azimuth_from_latitude.called)
        self.assertTrue(mock_calculate_table_length.called)
        self.assertTrue(mock_calculate_tables_per_row.called)

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_dc_field_width')
    @mock.patch('plantpredict.powerplant.PowerPlant._calculate_dc_field_length')
    def test_add_dc_field_dimension_calculator_helpers_called(self, mock_calculate_dc_field_length,
                                                              mock_calculate_dc_field_width):

        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        self.powerplant.add_dc_field(
            block_name=1,
            array_name=1,
            inverter_name="A",
            module_id=123,
            tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
            number_of_series_strings_wired_in_parallel=400,
            modules_high=4,
            modules_wired_in_series=10,
            post_to_post_spacing=1.0,
            tracking_backtracking_type=BacktrackingTypeEnum.BACKTRACKING,
            module_azimuth=180.0
        )

        self.assertTrue(mock_calculate_dc_field_length.called)
        self.assertTrue(mock_calculate_dc_field_width.called)

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_add_dc_field_fails_on_fixed_tilt_no_module_tilt(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        with self.assertRaises(ValueError):
            self.powerplant.add_dc_field(
                block_name=1,
                array_name=1,
                inverter_name="A",
                module_id=123,
                field_dc_power=800,
                tracking_type=TrackingTypeEnum.FIXED_TILT,
                modules_high=4,
                modules_wired_in_series=10,
                post_to_post_spacing=1.5
            )

    @mock.patch('plantpredict.plant_predict_entity.requests.get', new=mocked_requests.mocked_requests_get)
    @mock.patch('plantpredict.project.requests.get', new=mocked_requests.mocked_requests_get)
    def test_add_dc_field_fails_on_tracker_no_backtracking_type(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77)
        self._init_powerplant_structure()

        with self.assertRaises(ValueError):
            self.powerplant.add_dc_field(
                block_name=1,
                array_name=1,
                inverter_name="A",
                module_id=123,
                number_of_series_strings_wired_in_parallel=400,
                tracking_type=TrackingTypeEnum.HORIZONTAL_TRACKER,
                modules_high=4,
                modules_wired_in_series=10,
                post_to_post_spacing=1.5
            )

    def test_init(self):
        self._make_mocked_api()
        self.powerplant = PowerPlant(api=self.mocked_api, project_id=7, prediction_id=77, some_kwarg='kwarg')

        self.assertEqual(self.powerplant.project_id, 7)
        self.assertEqual(self.powerplant.prediction_id, 77)
        self.assertTrue(self.powerplant.use_cooling_temp)
        self.assertEqual(self.powerplant.power_factor, 1.0)
        self.assertEqual(self.powerplant.blocks, [])
        self.assertEqual(self.powerplant.transformers, [])
        self.assertEqual(self.powerplant.transmission_lines, [])
        self.assertEqual(self.powerplant.some_kwarg, 'kwarg')


if __name__ == '__main__':
    unittest.main()
